#!/usr/bin/python3
# Copyright (c) 2015-2020 SUSE Software Solutions Germany GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys
import re
import rpm
import io
import copy
import zlib
from stat import S_ISDIR

try:
    from lxml import etree
except ImportError:
    import cElementTree as etree

import osc.conf
import osc.core

# Disable xml tree limits as we have some large XML documents
# to parse. We don't have a DoS risk.
# http://lxml.de/FAQ.html#is-lxml-vulnerable-to-xml-bombs
hugeparser = etree.XMLParser(huge_tree=True)


class Package(object):

    def __init__(self, data, repo='SLE-12'):
        self.logger = repo.logger
        self.root = data
        self.repo = repo
        self.ns = self.root.nsmap
        self.ns['common'] = self.ns.pop(None, None)

    @property
    def name(self):
        return self.root.find('common:name', namespaces=self.ns).text

    @property
    def is_debug(self):
        if '-debuginfo' in self.name or '-debugsource' in self.name:
            return True

        return False

    @property
    def files(self):
        ns = self.repo.filelists.nsmap
        ns['filelists'] = ns.pop(None, None)
        x_path = "(filelists:package[@name='%s'])" % self.name

        try:
            result = self.repo.filelists.xpath(x_path, namespaces=ns)
        except TypeError as e:
            # FIX ME. This shouldn't happen but we have encountered TypeErrors
            # here which killed the script running on the bot.
            # We log and ignore for now, but need to investigate the root cause
            self.logger(e)
            self.logger('x_path: %s  namespace: %s' % (x_path, ns))

        files = []

        if not result:
            return files

        filelist = result[0]

        for f in filelist.xpath("filelists:file[not(@type='dir')]", namespaces=ns):
            files.append(f.text)
        return files


class Channel(object):
    def __init__(self, name, url, channeltype='product', logger=None):
        self.name = name
        self.url = url
        self.whitelist = None
        self.logger = logger
        self.type = channeltype

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __unicode__(self):
        return self.name

    def __hash__(self):
        return hash(self.url)

    def __eq__(self, other):
        return self.url == other.url

    def __lt__(self, other):
        return self.name < other.name

    def filename(self, ext=''):
        return '%s-%s%s' % (self.name, self.type, ext)

    def dump_filename(self):
        return self.filename('.dump')

    def channel_filename(self):
        return self.filename('.channel')

    def pickle_filename(self):
        return self.filename('.p')


class RepoData(Channel):

    def purge(self):
        try:
            del self._primary
        except Exception:
            pass
        try:
            del self._filelists
        except Exception:
            pass

    def get_repodata_db(self, db):
        repodata = self.repomd
        ns = repodata.nsmap
        ns['repo'] = ns.pop(None)
        if ns['repo'] is None:
            if self.logger:
                self.logger.error('Invalid repomd')
                self.logger.error('Returning empty data for %s' % db)
            return etree.fromstring('<dummy/>')

        x_path = 'repo:data[@type="{db}"]/repo:location'.format(db=db)

        if not repodata.xpath(x_path, namespaces=ns):
            return etree.fromstring('<dummy/>')

        url = repodata.xpath(x_path, namespaces=ns)[0].get('href', None)
        if url is not None:
            db_url = os.path.join(self.url, url)
        else:
            raise Exception('%s url not found!' % db)

        try:
            r = osc.core.http_GET(db_url).read()
        except Exception as e:
            if self.logger:
                self.logger.error(e)
                self.logger.error('Failed to retrieve %s' % db_url)
                self.logger.error('Returning empty data')
            return etree.fromstring('<dummy/>')

        d = zlib.decompressobj(16 + zlib.MAX_WBITS)
        data = []
        for chunk in io.BytesIO(r):
            data.append(d.decompress(chunk))

        xml = b''.join(data)

        return etree.fromstring(xml, parser=hugeparser)

    @property
    def primary(self):
        try:
            return self._primary
        except AttributeError:
            self._primary = self.get_repodata_db('primary')
            return self._primary

    @property
    def filelists(self):
        try:
            return self._filelists
        except AttributeError:
            self._filelists = self.get_repodata_db('filelists')
            return self._filelists

    @property
    def repomd(self):
        url = self.url + '/repodata/repomd.xml'
        try:
            r = osc.core.http_GET(url)
        except Exception as e:
            if self.logger:
                self.logger.error(e)
                self.logger.error('Failed to retrieve %s' % url)
                self.logger.error('Returning empty data')
            return etree.fromstring('<dummy/>')

        return etree.parse(r, parser=hugeparser).getroot()

    @property
    def packages(self):
        ns = self.primary.nsmap
        ns['common'] = ns.pop(None, None)
        if ns['common'] is None:
            return []
        return [Package(p, self) for p in self.primary.xpath('common:package', namespaces=ns)]


class FileListExtractor(object):
    def __init__(self, *args, **kwargs):
        self.ts = rpm.TransactionSet()
        self.ts.setVSFlags(rpm._RPMVSF_NOSIGNATURES)
        self.blacklist = None
        self.whitelist = None
        self.debug = False
        self.save_channels_files = False
        self.osc_initialized = False
        self.file_blacklist = None

    def compile_re(self, patterns):
        re_pattern = None
        if patterns:
            re_pattern = re.compile('(' + '|'.join(patterns) + ')')
        return re_pattern

    def set_blacklist(self, patterns):
        self.blacklist = self.compile_re(patterns)

    def set_whitelist(self, patterns):
        self.whitelist = self.compile_re(patterns)

    def set_file_blacklist(self, patterns):
        self.file_blacklist = self.compile_re(patterns)

    def _skip_file(self, fname):
        if self.file_blacklist and self.file_blacklist.match(fname):
            return True
        return False

    def _skip_pkg(self, pkgname):
        # Skip blacklisted packages. If we have a whitelist, then
        # only process files in the whitelist.
        if self.blacklist is not None and self.blacklist.match(pkgname):
            return True
        if self.whitelist is not None and not self.whitelist.match(pkgname):
            return True

        return False

    def merge(self, data1, data2):

        # http://stackoverflow.com/questions/11011756/is-there-any-pythonic-way-to-combine-two-dicts-adding-values-for-keys-that-appe
        # def combine_dicts(a, b, op=operator.add):
        #     return dict(a.items() + b.items() + [(k, op(a[k], b[k])) for k in set(b) & set(a)])

        def combine_dicts(a, b):
            return dict(list(a.items()) + list(b.items()) +  # noqa: W504
                        [(k, list(set.union(set(a[k]), set(b[k]))))
                            for k in set(b) & set(a)])

        pkgnames = data1['pkgnames']
        filenames = data1['filenames']

        pkgnames = pkgnames.union(data2['pkgnames'])
        filenames = combine_dicts(filenames, data2['filenames'])

        return {'filenames': filenames, 'pkgnames': pkgnames}

    def readRpmHeader(self, filename):
        """ Read an rpm header. """
        fd = os.open(filename, os.O_RDONLY)
        h = self.readRpmHeaderFD(fd)
        os.close(fd)
        return h

    def readRpmHeaderFD(self, fd):
        h = None
        try:
            h = self.ts.hdrFromFdno(fd)
        except rpm.error as e:
            if str(e) == "public key not available":
                print(str(e))
            if str(e) == "public key not trusted":
                print(str(e))
            if str(e) == "error reading package header":
                print(str(e))
            h = None
        return h

    def _fetchcpioheaders(self, apiurl, tofetch):
        # http://stackoverflow.com/questions/312443/how-do-you-split-a-list-into-evenly-sized-chunks-in-python
        def chunks(l, n):
            """ Yield successive n-sized chunks from l.
            """
            for i in range(0, len(l), n):
                yield l[i:i + n]

        from osc.util.cpio import CpioRead

        # fetch rpm headers for binaries from source projects.
        for repo in tofetch.keys():
            for binaries in chunks(list(tofetch[repo]), 50):
                u = osc.core.makeurl(apiurl, ['build', repo[0], repo[1], repo[2], '_repository'],
                                     ['view=cpioheaders'] + ['binary=%s' % i for i in binaries])
                r = osc.core.http_GET(u)
                from tempfile import NamedTemporaryFile
                tmpfile = NamedTemporaryFile(prefix="cpio-", delete=False)
                for chunk in r:
                    tmpfile.write(chunk)
                tmpfile.close()
                cpio = CpioRead(tmpfile.name)
                cpio.read()
                for ch in cpio:
                    # some binary files may not exist if no update was
                    # released yet so ignore.
                    if ch.filename == '.errors':
                        continue
                    # the filehandle in the cpio archive is private so
                    # open it again
                    with open(tmpfile.name, 'rb') as fh:
                        fh.seek(ch.dataoff, os.SEEK_SET)
                        h = self.readRpmHeaderFD(fh)
                        if h is None:
                            self.logger.warn("failed to read rpm header for %s" % ch.filename)
                        else:
                            yield h
                os.unlink(tmpfile.name)

    def _init_osc(self):
        if self.osc_initialized:
            return
        self.osc_initialized = True
        osc.conf.get_config(override_apiurl='https://api.suse.de/')
        if (self.debug):
            osc.conf.config['debug'] = 1

    def _get_channel_xml(self, channel):
        if channel.url.startswith('file://'):
            fn = channel.url.split('://')[1]
            try:
                with open(fn, 'rb') as f:
                    root = etree.parse(f).getroot()
            except Exception as e:
                self.logger.error("Failed to open channel file: %s" % fn)
                self.logger.error(e)
                return None
        else:
            try:
                self._init_osc()
                osc.conf.config['apiurl']
                r = osc.core.http_GET(channel.url)
                root = etree.parse(r).getroot()
            except Exception as e:
                self.logger.warning("Failed to read channel data: %s" % channel.url)
                self.logger.debug(e)
                if 'SUSE:Channels/' in channel.url:
                    eol = copy.copy(channel)
                    eol.url = eol.url.replace('SUSE:Channels', 'SUSE:Channels:EOL')
                    self.logger.info('Trying: %s' % eol.url)
                    return self._get_channel_xml(eol)
                return None

        return root

    def _parse_channel_xml(self, root, callback, whitelist=None):
        self._init_osc()
        tofetch = dict()
        self.set_whitelist(whitelist)
        for binaries in root.findall('binaries'):
            repo = (binaries.attrib['project'], binaries.attrib['repository'], binaries.attrib['arch'])
            for node in binaries.findall('binary'):
                name = node.attrib['name']
                if name.endswith('-debuginfo') or name.endswith('-debugsource'):
                    continue
                if self._skip_pkg(name):
                    continue
                tofetch.setdefault(repo, set()).add(name)

        for h in self._fetchcpioheaders(osc.conf.config['apiurl'], tofetch):
            callback(h)

    def readFileLists(self, args):

        filenames = dict()
        pkgnames = set()

        def _getdata(h):
            # don't care about debuginfo
            if h['name'].endswith('-debuginfo') or h['name'].endswith('-debugsource'):
                return
            pkgnames.add(h['name'])
            for fn, mode in zip(h['filenames'], h['filemodes']):
                if S_ISDIR(mode):
                    continue
                if self._skip_file(fn):
                    continue
                filenames.setdefault(fn, set()).add(h['name'])

        for arg in args:
            if isinstance(arg, RepoData):
                self._init_osc()
                self.set_whitelist(arg.whitelist)
                for pkg in arg.packages:
                    if pkg.is_debug:
                        continue
                    if self._skip_pkg(pkg.name):
                        continue
                    pkgnames.add(pkg.name)
                    for fn in pkg.files:
                        if self._skip_file(fn):
                            continue
                        filenames.setdefault(fn, set()).add(pkg.name)

                # Done processing repodata
                # Remove cached data to conserve on memory usage
                arg.purge()

            elif isinstance(arg, Channel):
                self.logger.info('Reading channel: %s' % arg.name)
                root = self._get_channel_xml(arg)
                if root is None:
                    continue
                if self.save_channels_files:
                    with open(arg.channel_filename(), 'wb') as f:
                        f.write(etree.tostring(root))
                self._parse_channel_xml(root, _getdata, arg.whitelist)

            elif arg.endswith('.iso'):
                self.logger.info('Reading ISO: %s' % arg)
                import pycdio
                import iso9660
                iso = iso9660.ISO9660.IFS(source=arg)
                fd = os.open(arg, os.O_RDONLY)

                if not iso.is_open() or fd is None:
                    raise Exception("Could not open %s as an ISO-9660 image." % arg)

                for path in ['/suse/x86_64/', '/suse/noarch']:
                    file_stats = iso.readdir(path)
                    if file_stats is None:
                        continue
                    for stat in file_stats:
                        filename = stat[0]
                        LSN = stat[1]

                        if (filename.endswith('.rpm')):
                            os.lseek(fd, LSN * pycdio.ISO_BLOCKSIZE, io.SEEK_SET)
                            h = self.ts.hdrFromFdno(fd)
                            _getdata(h)

                os.close(fd)

            elif os.path.isfile(arg) and arg.endswith('.rpm'):
                self.logger.info('Reading package %s' % arg)
                h = self.readRpmHeader(arg)
                _getdata(h)
            elif os.path.isdir(arg):
                for root, dirs, files in os.walk(arg):
                    self.readFileLists([os.path.join(root, f) for f in files])
            else:
                print("don't know what to do with %s" % arg, file=sys.stderr)

        # Convert sets to lists in filenames dict.
        filenames = dict([(k, list(filenames[k])) for k in filenames])
        return {'filenames': filenames, 'pkgnames': pkgnames}


# vim: sw=4 et
