#!/usr/bin/python3
# Copyright (c) 2015-2020 SUSE Software Solutions Germany GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click
import logging
import os
import pickle as pickle
import psutil
import re
import shelve
import signal
import subprocess
import sys
import time
import traceback
import yaml
from collections import namedtuple
from pprint import pformat
from .FileListExtractor import FileListExtractor, Channel, RepoData

USER_CONFIG_DIR = os.environ.get('XDG_CONFIG_HOME', os.path.expanduser('~/.config'))
SYS_CONFIG_DIRS = os.environ.get('XDG_CONFIG_DIRS', '/etc/xdg').split(':')
CONFIG_DIRS = [USER_CONFIG_DIR] + SYS_CONFIG_DIRS
CONFIG_DIRS = [os.path.join(d, 'rpmlint-backports-tool') for d in CONFIG_DIRS]
BLACKLIST = ['java-.*-ibm.*']
DEFAULT_DBFILE = 'SLE-file-package-map.db'
module_re = re.compile(".*(sle-module|sle-manager)", flags=re.IGNORECASE)

PickleResult = namedtuple('PickleResult', 'newpkgs newfiles missing')

logging.basicConfig()
logger = logging.getLogger('rpmlint-backports-tool')


def memory_usage_psutil():
    '''Return memory usage in megabytes'''
    process = psutil.Process(os.getpid())

    # First try current api and fall back to old api on error
    try:
        mem = process.memory_info()[0] / float(2 ** 20)
    except AttributeError:
        mem = process.get_memory_info()[0] / float(2 ** 20)
    return mem


def service_pack(name, SP=None, url=False):
    if name is None:
        return None
    if (not SP or module_re.match(name)):
        return name
    if SP == 'SP0':
        return name
    if not url:
        return '%s-%s' % (name, SP)
    return re.sub(r'([_\-]12)([_\-:])', r'\1-%s\2' % SP, name)


class Tool():
    def __init__(self):
        self.ex = FileListExtractor()
        self.ex.save_channels_files = True
        self._channels = None
        self._channels_by_arch = {}
        self._config = None
        self._dryrun = None
        self._servicepack = None

    def reset(self):
        """ Clear any cached data """
        self._channels = None

    @property
    def config(self):
        if self._config is None:
            raise Exception("No config file loaded")
        return self._config

    @config.setter
    def config(self, data):
        self._config = data

    @property
    def channels(self):
        if self._channels is None:
            self._channels = self._get_channels()
        return self._channels

    @property
    def channels_by_arch(self):
        if self.channels is not None:
            return self._channels_by_arch

    @property
    def servicepacks(self):
        if self._servicepack:
            return [self._servicepack]
        return self.config.get('servicepacks', None)

    def _get_channels(self):
        channels = []
        chan_url = {}
        chan_url['product'] = self.config.get('product_channels_url', None)
        chan_url['update'] = self.config.get('update_channels_url', None)
        chan_url['local'] = self.config.get('local_channels_url', '{name}')
        channel_configs = self.config.get('channels', None)
        channel_schema = self.config.get('channel_schema', '_channel')

        if channel_configs is None:
            raise Exception("No channels in config")
        if chan_url['product'] is None:
            raise Exception("No product_channels_url in config")
        if chan_url['update'] is None:
            raise Exception("No update_channels_url in config")

        for ID, config in channel_configs.items():
            archs = config.get('archs', ['x86_64'])
            for channeltype in ['product', 'update', 'local']:
                name = config.get(channeltype, None)
                if name is None:
                    continue
                servicepacks = self.servicepacks or ['SP0']
                if module_re.match(name) or channeltype == 'local':
                    servicepacks = ['SP0']
                for servicepack in servicepacks:
                    if '@' in servicepack:
                        servicepack, snapshot = servicepack.split('@', 1)
                    else:
                        snapshot = ''
                    for arch in archs:
                        url = chan_url[channeltype].format(name=name,
                                                           arch=arch.replace('+', '_'),
                                                           snapshot=snapshot
                                                           )
                        url = service_pack(url, SP=servicepack, url=True)
                        sp_name = service_pack(os.path.split(name)[1], SP=servicepack)
                        sp_name = '%s-%s' % (sp_name, arch.replace('+', '_'))
                        if channel_schema == 'repomd':
                            newchannel = RepoData(sp_name, url, channeltype, logger=self.logger)
                        else:
                            newchannel = Channel(sp_name, url, channeltype, logger=self.logger)
                        newchannel.id = ID
                        newchannel.exclude = config.get('exclude', False)
                        newchannel.whitelist = config.get('whitelist', None)

                        # Multi arch channels have the multi archs in one string
                        # separated by a '+'. The packages in that channel need
                        # to be added to all arch channels.
                        if '+' in arch:
                            splitarchs = arch.split('+')
                        else:
                            splitarchs = [arch]
                        for arch in splitarchs:
                            self._channels_by_arch.setdefault(arch, []).append(newchannel)
                        channels.append(newchannel)

        return channels

    def filter_channels(self, channel_names):
        if channel_names:
            names = set(channel_names)
            return [c for c in self.channels if c.name in names]

        return self.channels

    def load_config(self, configfile):
        def _load(path):
            self.logger.debug("Loading %s" % path)
            try:
                with open(path) as f:
                    self.config = yaml.load(f, Loader=yaml.CLoader)
                self.logger.info("Loaded config file: %s" % path)
                return True
            except Exception as e:
                self.logger.debug("Failed to load %s. %s" % (path, e))
                return False

        # First try path exactly as specified
        if os.path.isfile(os.path.abspath(configfile)) and _load(configfile):
            return True

        # Search for config file in standard locations
        # Load first match and return
        filenames = [configfile, configfile + '.conf', configfile + '.config']
        for path in CONFIG_DIRS:
            for fname in filenames:
                file_path = os.path.abspath(os.path.join(path, fname))
                self.logger.debug("Checking for %s" % file_path)
                if not os.path.isfile(file_path):
                    continue
                if _load(file_path):
                    return True

        raise Exception('Failed to find and load config: %s' % configfile)

    def pickle(self, channels_to_process, force=None):
        if not force:
            force = self.config.get('force', False)

        self.ex.set_blacklist(BLACKLIST)  # What is this for?
        self.ex.set_file_blacklist(self.config.get('file_blacklist', None))
        stats = {}

        def process(channel):
            self.logger.debug("processing %s", channel.name)
            fn = channel.pickle_filename()
            missing = []
            newpkgs = []
            newfiles = []
            olddata = None
            if os.path.exists(fn):
                if channel.type == 'product' and not force:
                    self.logger.info('Channel type is product - skipping refresh.')
                    return None
                with open(fn, 'rb') as f:
                    try:
                        olddata = pickle.load(f)
                    except Exception as e:
                        logger.warning(e)
                        logger.warning('Failed to load existing pickle file %s' % fn)
                        olddata = None

            data = self.ex.readFileLists([channel])

            # FileListExtractor will return empty package list if reading
            # repodata fails. Handle zero lists as error.
            if len(data['pkgnames']) == 0:
                if olddata is not None and len(olddata['pkgnames']) > 0:
                    self.logger.warning('Channel has no packages! Skipping...')
                    return None

            stats['pkgs_processed'] = len(data['pkgnames'])
            stats['total_files'] = len(data['filenames'])
            if olddata is not None:
                missing = olddata['pkgnames'] - data['pkgnames']
                newpkgs = data['pkgnames'] - olddata['pkgnames']
                newfiles = set(data['filenames'].keys()) - set(olddata['filenames'].keys())
            else:
                newpkgs = ['%s packages from %s' % (len(data['pkgnames']), channel.name)]

            if newfiles or newpkgs or missing:
                with open(fn, 'wb') as f:
                    pickle.dump(data, f, pickle.HIGHEST_PROTOCOL)
                with open(channel.dump_filename(), 'wb') as f:
                    f.write(pformat(data).encode('utf-8'))

            return PickleResult(newpkgs=newpkgs, newfiles=newfiles, missing=missing)

        changed = set()
        for arch, channels in self.channels_by_arch.items():
            for channel in channels:
                if channel not in channels_to_process:
                    continue
                start = time.time()

                title = '{} {}'.format(channel.name, channel.type)
                dashes = '-' * int((70 - (len(title) + 2)) / 2)
                self.logger.info('%s %s %s' % (dashes, title, dashes))
                self.logger.info('repo: %s' % (channel.url))

                result = process(channel)
                if result is None:
                    self.logger.info('-' * 70)
                    continue

                self.logger.info('Processed %.2f seconds' % (time.time() - start))
                self.logger.info('Current memory usage: %.1fM' % memory_usage_psutil())
                self.logger.info('Scanned %s packages' % stats.get('pkgs_processed', 0))
                self.logger.info('Collected %s filenames' % stats.get('total_files', 0))

                if result.missing:
                    self.logger.warning("  vanished packages: %s" % ', '.join(result.missing))
                    changed.add(arch)
                if result.newpkgs:
                    self.logger.info("  new packages: %s" % ', '.join(result.newpkgs))
                    changed.add(arch)
                if result.newfiles:
                    self.logger.info("  %s new files added" % len(result.newfiles))
                    changed.add(arch)
                if not (result.newpkgs or result.newfiles or result.missing):
                    self.logger.info("  no change")
                self.logger.info('-' * 70)

        return changed

    def merge(self, filename, files, shelve=False, by_arch=False, archs=None):

        dbfiles = []

        if archs is None:
            archs = self.channels_by_arch.keys()

        # If by_arch is set, then we call ourselves for each arch, merging
        # only the channels for that arch in a separate db file set
        if by_arch:
            for arch, channels in self.channels_by_arch.items():
                if arch not in archs:
                    continue
                pfiles = [c.pickle_filename() for c in sorted(channels) if not c.exclude]
                pfiles = [f for f in pfiles if os.path.exists(f)]
                dbfile = '-{}.'.format(arch).join(filename.split('.'))
                self.merge(dbfile, pfiles, shelve)
                dbfiles.append(dbfile)

            return dbfiles

        dbfiles = [filename]
        if not files:
            files = [c.pickle_filename() for c in sorted(self.channels) if not c.exclude]
            files = [f for f in files if os.path.exists(f)]

        res = {
            'filenames': dict(),
            'pkgnames': set(),
        }

        for fn in files:
            self.logger.info("merging %s" % fn)
            with open(fn, 'rb') as f:
                if not os.path.exists(fn):
                    logger.warning('%s not found!' % fn)
                    continue
                try:
                    data = pickle.load(f)
                    res = self.ex.merge(res, data)
                except Exception as e:
                    logger.error(e)
                    logger.error('Failed to merge %s' % fn)
                    raise(e)

        filename = filename
        tmpfn = filename + '.new'
        if shelve:
            import shelve
            # For the shelve backend db, Python2 uses bsddb by default,
            # Python3 does not support bsddb. The gdbm format is endian
            # sensitive which causes issues when trying to use a db generated
            # on a machine with a different endianness as the system running
            # the rpmlint checks at build time.
            # So we explictly use dumbdbm for compatiblity with python 2 and
            # 3 (i.e. SLES 12 and SLES 15) as well as little and big
            # endian systems. Also ensure same protocol (2) is used.

            # Trick anydbm into using dumbdbm by creating empty .dat and .dir
            # files so that shelve.open thinks we are opening a dumbdbm based
            # db

            for ext in ['.dat', '.dir']:
                with open(tmpfn + ext, 'w') as f:
                    pass

            d = shelve.open(tmpfn, protocol=2)
            d.update(res)
            d.close()

        else:
            with open(tmpfn, 'wb') as f:
                pickle.dump(res, f, pickle.HIGHEST_PROTOCOL)

        # dbm dumb creates 3 files filename.dat and filename.dir filename.bak
        # We need the .dir and .dat files
        for ext in ['.dat', '.dir']:
            os.rename(tmpfn + ext, filename + ext)

        os.remove(tmpfn + '.bak')

        self.logger.debug('Current memory usage: %s' % memory_usage_psutil())

        return dbfiles

    def upload(self, dbfile):
        import osc.core
        if not os.path.exists(dbfile):
            raise Exception("Can't find %s. Run pickle and merge" % dbfile)
        self.logger.info("uploading")
        self.ex._init_osc()
        apiurl = self.config.get('rpmling_apiurl', 'https://api.opensuse.org')
        prj = self.config.get('rpmlint_prj', None)
        if prj is None:
            raise Exception("No project defined for obs update!")
        pkg = 'rpmlint-backports-data'
        u = osc.core.makeurl(apiurl, ['source', prj, pkg, dbfile], {})
        self.logger.info('Preparing to upload %s to %s' % (dbfile, u))
        if self._dryrun:
            self.logger.info('Dry run - upload aborted')
            return
        r = osc.core.http_PUT(u, file=dbfile)
        self.logger.info(r.read())


@click.group()
@click.option("--dry", is_flag=True, help="dry run")
@click.option("--debug", is_flag=True, help="debug output")
@click.option("--verbose", is_flag=True, help="verbose")
@click.option("--servicepack", help="service pack", default=None)
@click.option("--config", help="config file", default=None)
@click.option("--workpath", help="work path", default='.')
@click.pass_context
def cli(ctx, dry, debug, verbose, servicepack, config, workpath):

    ctx.ensure_object(dict)
    tool = Tool()
    if debug:
        logger.setLevel(logging.DEBUG)
        tool.ex.debug = True
    elif verbose:
        logger.setLevel(logging.INFO)

    tool.logger = logger
    tool.ex.logger = logger

    if config is not None:
        tool.load_config(config)

    if workpath is not None:
        try:
            os.chdir(os.path.abspath(workpath))
        except Exception as e:
            raise Exception("Failed to switch to workpath %s: %s" % (workpath, e))

    tool._servicepack = servicepack
    tool._dryrun = dry

    ctx.obj['TOOL'] = tool


@cli.command()
@click.argument('channel_names', nargs=-1)
@click.pass_context
def list(ctx, channel_names):
    """list channels
    """
    tool = ctx.obj['TOOL']
    channels = tool.filter_channels(channel_names)
    for c in sorted(channels):
        click.echo("%s\n  %s\n  %s" % (c.name, c.url, c.channel_filename()))


@cli.command('pickle')
@click.option("-f", "--force", is_flag=True, help='force refresh of "product repos". By default only update repos are refreshed')
@click.argument('channel_names', nargs=-1)
@click.pass_context
def do_pickle(ctx, force, channel_names):
    """Generate pickle for channels
    """

    tool = ctx.obj['TOOL']
    channels = tool.filter_channels(channel_names)
    tool.pickle(sorted(channels), force=force)


@cli.command()
@click.option("-o", "--output", 'filename', required=True, help="output to FILE")
@click.option("--shelve", is_flag=True, help="save as shelve file")
@click.option("--by-arch", is_flag=True, help="save one database per arch")
@click.argument('files', nargs=-1)
@click.pass_context
def merge(ctx, filename, shelve, by_arch, files):
    """Merge .p files after call to pickle
    """

    tool = ctx.obj['TOOL']
    return '\n'.join(tool.merge(filename, files, shelve=shelve, by_arch=by_arch))


def db_pretty_print(data, only=None):
    def print_packages(data):
        click.echo('Packages')
        click.echo('========')
        for pkgname in data['pkgnames']:
            click.echo('  %s' % pkgname)

    def print_files(data):
        click.echo('Package Files')
        click.echo('=============')
        for fn, pkgs in data['filenames'].items():
            click.echo('  %s' % fn)
            pkgs = ', '.join(pkgs)
            click.echo('    provided by %s' % pkgs)
            click.echo()

    if only == 'packages':
        print_packages(data)
    elif only == 'files':
        print_files(data)
    else:
        print_packages(data)
        click.echo()
        print_files(data)


@cli.command()
@click.option('--packages', 'only', flag_value='packages', help='only print package name list')
@click.option('--files', 'only', flag_value='files', help='only print file list')
@click.argument('datafile')
def dump_database(datafile, only):
    """Pretty print the data from the specified pickle or shelve file
       that has been generated using this tool via pickle or merge commands.
    """
    try:
        with shelve.open(datafile, 'r') as d:
            db_pretty_print(d, only=only)
    except Exception:
        try:
            with open(datafile, 'rb') as f:
                db_pretty_print(pickle.load(f), only=only)
        except Exception as e:
            logger.error(e)
            logger.error('Failed to open pickle or shelve file')


@cli.command()
@click.argument('dbfiles', nargs=-1)
@click.pass_context
def update_obs(ctx, dbfiles):
    """Upload DBFILES to rpmlint-backports-data package
       DBFILES must be specified WITHOUT .dat or .dir extension.
    """
    tool = ctx.obj['TOOL']
    for dbfile in dbfiles:
        for ext in ['dat', 'dir']:
            fn = dbfile + '.' + ext
            tool.upload(fn)


@cli.command()
@click.option('-n', '--interval', help="periodic interval in minutes")
@click.option('--git-commit', is_flag=True, help="commit changes to git")
@click.option("-f", "--force", is_flag=True, help="force refresh of product repos (use for beta releases)")
@click.pass_context
def run_bot(ctx, interval, git_commit, force):
    """Update files, upload
    """
    tool = ctx.obj['TOOL']

    class ExTimeout(Exception):
        """raised on timeout"""

    if interval:
        def alarm_called(nr, frame):
            raise ExTimeout()
        signal.signal(signal.SIGALRM, alarm_called)

    while True:
        tool.reset()
        start = time.time()
        try:
            channels = sorted(tool.channels)
            changed = tool._pickle(channels, force=force)
            if changed:
                if git_commit:
                    try:
                        subprocess.check_output(['git', 'init'])
                        for c in channels:
                            subprocess.check_output(['git', 'add', c.dump_filename()])
                        subprocess.check_output(['git', 'commit', '-m', 'update'])
                    except subprocess.CalledProcessError as e:
                        logger.warning("### git ERROR: %s" % e)
                logger.info("merging")
                dbfiles = tool.merge(DEFAULT_DBFILE, None, shelve=True, by_arch=True, archs=changed)
                for dbfile in dbfiles:
                    for ext in ['dat', 'dir']:
                        fn = dbfile + '.' + ext
                        tool.upload(fn)
        except Exception as e:
            logger.error("### ERROR: %s" % e)
            logger.info(traceback.format_exc())

        logger.info('====================================================================')
        logger.info('Completed in %s seconds' % (time.time() - start))
        logger.info('Current memory usage: %.1fM' % memory_usage_psutil())
        logger.info('====================================================================')

        if interval:
            logger.info("sleeping %d minutes. Press enter to check now ..." % interval)
            signal.alarm(interval * 60)
            try:
                input()
            except ExTimeout:
                pass
            except EOFError:
                # no tty available, disable alarm and sleep
                signal.alarm(0)
                time.sleep(interval * 60)
            continue
        break


if __name__ == "__main__":
    try:
        sys.exit(cli())
    except Exception as e:
        logger.error('%s. exiting...' % e, exc_info=True)
        sys.exit(1)

# vim: sw=4 et
